//
// Copyright (C) 2017 Kevin Thornton <krthornt@uci.edu>
//
// This file is part of fwdpy11.
//
// fwdpy11 is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// fwdpy11 is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with fwdpy11.  If not, see <http://www.gnu.org/licenses/>.
//

#ifndef FWDPY11_DISCRETE_DEMOGRAPY_SIMULATION_HPP
#define FWDPY11_DISCRETE_DEMOGRAPY_SIMULATION_HPP

#include "constants.hpp"
#include "simulation/get_max_number_of_demes.hpp"
#include "simulation/apply_mass_migrations.hpp"
#include "simulation/deme_properties.hpp"
#include "simulation/multideme_fitness_lookups.hpp"
#include "simulation/migration_lookup.hpp"
#include "simulation/build_migration_lookup.hpp"
#include "simulation/functions.hpp"
#include "simulation/pick_parents.hpp"
#include "simulation/demographic_model_state.hpp"
#include "simulation/update_demography_manager.hpp"
#include "simulation/set_model_state.hpp"

#endif
