#ifndef FWDPY11_EVOLVE_REMOVE_EXTINCT_GENOMES_HPP
#define FWDPY11_EVOLVE_REMOVE_EXTINCT_GENOMES_HPP

#include <fwdpy11/types/DiploidPopulation.hpp>

void
remove_extinct_genomes(fwdpy11::DiploidPopulation& pop);

#endif

