#ifndef __FWDPP_TAGS_TAGS_HPP__
#define __FWDPP_TAGS_TAGS_HPP__

namespace fwdpp
{
    namespace tags
    {
        //! Empty struct acts like a dispatch tag for fwdpp::haploid_genome_base 
        struct standard_haploid_genome
        {
        };
    }
}

#endif
