#ifndef FWDPP_TS_MARGINAL_TREE_FUNCTIONS_HPP__
#define FWDPP_TS_MARGINAL_TREE_FUNCTIONS_HPP__

#include "marginal_tree_functions/roots.hpp"
#include "marginal_tree_functions/children.hpp"
#include "marginal_tree_functions/nodes.hpp"
#include "marginal_tree_functions/samples.hpp"

#endif
